from os.path import join, dirname, exists, basename, splitext
from os import mkdir
import numpy as np
from tqdm.contrib.concurrent import process_map
from itertools import repeat
from functools import partial
from videochef.io import videoReader, videoWriter
from videochef.util import gen_batch_sequence, count_frames


def parallel_proc_frame(vid_path, frame_batch, reporter_val, writer_path, analysis_func=None):
    with videoReader(vid_path, np.array(frame_batch), reporter_val) as vid, videoWriter(writer_path) as writer:
        for frame in vid:
            writer.append(analysis_func(frame))
            

def video_chef(func, path_to_vid, max_workers=3, frame_batch_size=500, vid_read_reporter=False, tmp_dir=None, proc_suffix='_PROC'):
    nframes = count_frames(path_to_vid)
    batch_seq = gen_batch_sequence(nframes, frame_batch_size, 0)
    vid_name, vid_ext = splitext(basename(path_to_vid))
    vid_dir = dirname(path_to_vid)

    if tmp_dir is None:
        tmp_dir = join(vid_dir, 'tmp')
        mkdir(tmp_dir)
    else:
        assert exists(tmp_dir)

    print('Processing in parallel...')
    parallel_writer_names = [join(tmp_dir, f'proc_{i}.avi') for i in range(len(batch_seq))]
    if vid_read_reporter:
        reporter_vals = (i for i in range(len(batch_seq)))
    else:
        reporter_vals = (None for i in range(len(batch_seq)))
    process_map(partial(parallel_proc_frame, analysis_func=func),
                repeat(path_to_vid),
                batch_seq,
                reporter_vals,
                parallel_writer_names,
                chunksize=1,
                max_workers=max_workers)

    print('Stitching parallel videos')
    stitched_vid_name = join(tmp_dir, vid_name + proc_suffix + vid_ext)
    with videoWriter(stitched_vid_name) as stitched_vid:
        for i, vid_name in enumerate(parallel_writer_names):
            print(f'Stitching video {i}')
            with videoReader(vid_name) as cheffed_vid:
                for frame in cheffed_vid:
                    stitched_vid.append(frame)

    return stitched_vid_name



import numpy as np
import av
import subprocess
import datetime


class videoWriter():
    def __init__(self, file_name, **ffmpeg_options):
        self.pipe = None
        self.file_name = file_name
        self.ffmpeg_options = ffmpeg_options
        
    def __enter__(self):
        return self
    
    def __exit__(self, type, value, traceback):
        if self.pipe is not None:
            self.pipe.stdin.close()
        
    def append(self, frames):
        if len(frames.shape)==2: frames = frames[None]
        self.pipe = write_frames(self.file_name, frames, pipe=self.pipe, **self.ffmpeg_options)

class videoReader():
    def __init__(self, file_name, frame_ixs=None, reporter_val=None):
        self.file_name = file_name
        self.frame_ixs = frame_ixs
        self.reporter_val = reporter_val

        if frame_ixs is not None:
            self.first_frame_ix = int(np.min(frame_ixs))
            self.final_frame_ix = int(np.max(frame_ixs))
        else:
            self.first_frame_ix = 0
            self.final_frame_ix = None
        
        if self.reporter_val is not None:
            self.reporter_val = int(reporter_val)

    def __enter__(self):
        self.reader = av.open(self.file_name, 'r')
        self.reader.streams.video[0].thread_type = "AUTO"
        
        # Create frame mask
        frame_mask = np.zeros(self.reader.streams.video[0].frames)
        if self.frame_ixs is not None:
            assert self.frame_ixs.max() < len(frame_mask), 'frame_ixs exceeds the video length'
            frame_mask[self.frame_ixs.astype(int)] = 1
        else: 
            frame_mask[:] = 1

        # Seek to first frame to use. When found, pass it to the frame_gen func for use as the first frame.
        if self.first_frame_ix != 0:
            self.first_frame_to_yield = self._precise_seek(self.first_frame_ix)
        else:
            self.first_frame_to_yield = None

        # Return a generator object that will decode the frames into np arrays
        return self.frame_gen(self.reader.decode(video=0), frame_mask)

    def _plain_frame_gen(self):
        for packet in self.reader.demux(self.reader.streams.video[0]):
            for frame in packet.decode():
                yield frame

    def _precise_seek(self, frame_num_to_seek):
        fps = self.reader.streams.video[0].average_rate
        time_base = self.reader.streams.video[0].time_base
        if not fps == 1/time_base:
            raise ValueError('Videoreader expects timebase to be 1 per frame!')
        self.reader.seek(frame_num_to_seek, stream=self.reader.streams.video[0])
        for frame in self._plain_frame_gen():
            if frame.pts >= frame_num_to_seek:
                return frame

    def frame_gen(self, reader, frame_mask):

        # Handle initial boundary condition from precise seeking
        if self.first_frame_to_yield is not None:
            self.ix_offset = 1
            yield self.first_frame_to_yield.to_ndarray()
        else:
            self.ix_offset = 0

        # Then yield subsequent frames
        for relative_ix,frame in enumerate(reader):
            ix = relative_ix + self.first_frame_ix + self.ix_offset # since enumerate always starts at 0, even if frame is not 0th frame
            if self.reporter_val: print(f'read frame {ix} from reader {self.reporter_val}')
            if self.final_frame_ix is not None and (ix > self.final_frame_ix): break  # short circuit to speed up multiprocessing
            if frame_mask[ix]: yield frame.to_ndarray()
            

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.reader.close()            
                
def write_frames(filename, frames, 
                 threads=6, fps=30, crf=10,
                 pixel_format='gray8', codec='ffv1',
                 pipe=None, slices=24, slicecrc=1):
    
    frame_size = '{0:d}x{1:d}'.format(frames.shape[2], frames.shape[1])
    command = ['ffmpeg',
               '-y',
               '-loglevel', 'fatal',
               '-framerate', str(fps),
               '-f', 'rawvideo',
               '-s', frame_size,
               '-pix_fmt', pixel_format,
               '-i', '-',
               '-an',
               '-crf',str(crf),
               '-vcodec', codec,
               '-preset', 'ultrafast',
               '-threads', str(threads),
               '-slices', str(slices),
               '-slicecrc', str(slicecrc),
               '-r', str(fps),
               filename]

    if not pipe: pipe = subprocess.Popen(command, stdin=subprocess.PIPE, stderr=subprocess.PIPE)
    dtype = np.uint16 if pixel_format.startswith('gray16') else np.uint8
    for i in range(frames.shape[0]): pipe.stdin.write(frames[i,:,:].astype(dtype).tobytes())
    return pipe


def read_frames(filename, frames, threads=6, fps=30, frames_is_timestamp=False,
                pixel_format='gray16', frame_size=(640,576),
                slices=24, slicecrc=1, get_cmd=False):
    """
    Reads in frames from the .mp4/.avi file using a pipe from ffmpeg.
    Args:
        filename (str): filename to get frames from
        frames (list or 1d numpy array): list of frames to grab
        threads (int): number of threads to use for decode
        fps (int): frame rate of camera in Hz
        pixel_format (str): ffmpeg pixel format of data
        frame_size (str): wxh frame size in pixels
        slices (int): number of slices to use for decode
        slicecrc (int): check integrity of slices
    Returns:
        3d numpy array:  frames x h x w
    """

    if frames_is_timestamp: start_time = str(datetime.timedelta(seconds=frames[0]))
    else: start_time = str(datetime.timedelta(seconds=frames[0]/fps))
    
    command = [
        'ffmpeg',
        '-loglevel', 'fatal',
        '-vsync','0',
        '-ss', start_time,
        '-i', filename,
        '-vframes', str(len(frames)),
        '-f', 'image2pipe',
        '-s', '{:d}x{:d}'.format(frame_size[0], frame_size[1]),
        '-pix_fmt', pixel_format,
        '-threads', str(threads),
        '-slices', str(slices),
        '-slicecrc', str(slicecrc),
        '-vcodec', 'rawvideo',
        '-'
    ]

    if get_cmd:
        return command
    
  
    pipe = subprocess.Popen(command, stderr=subprocess.PIPE, stdout=subprocess.PIPE)
    out, err = pipe.communicate()
    if(err):
        print('error', err)
        return None
    
    dtype = ('uint16' if '16' in pixel_format else 'uint8')
    video = np.frombuffer(out, dtype=dtype).reshape((len(frames), frame_size[1], frame_size[0]))
    return video


import numpy as np
import av

def gen_batch_sequence(nframes, chunk_size, overlap, offset=0):
    '''
    Generates batches used to chunk videos prior to extraction.
    Parameters
    ----------
    nframes (int): total number of frames
    chunk_size (int): desired chunk size
    overlap (int): number of overlapping frames
    offset (int): frame offset
    Returns
    -------
    Yields list of batches
    '''

    seq = range(offset, nframes)
    out = []
    for i in range(0, len(seq) - overlap, chunk_size - overlap):
        out.append(seq[i:i + chunk_size])
    return out


def dummy_process(frame):
    # Normalize the frame's pixel vals around half max brightness (just something stupid that isn't instant)
    avg = np.mean(frame)
    std = np.std(frame)
    frame = ((frame - avg)/std + 255/2).astype('uint8')
    return frame


def count_frames(file_name):
    with av.open(file_name, 'r') as reader:
        return reader.streams.video[0].frames          



import numpy as np
import av
import os
from os.path import join, exists
import matplotlib.pyplot as plt
from sklearn.datasets import load_digits
import time
import cv2
from tqdm import tqdm
from tqdm.contrib.concurrent import process_map
from imageio import get_reader
from itertools import repeat, count
import src.videochef.io as io

import imageio as iio
import pdb


# def parallel_dummy_process_frame(fname, frame_ixs, i):
#     print(f'starting reader {i}')
#     with io.videoReader(fname, frame_ixs) as vid:
#         for iFrame, frame in enumerate(vid):
#             if iFrame % 100 == 0:
#                 print(f'Decoded frame from reader {i}')
#             _ = dummy_process_frame(frame)


# def chunksize_test_func(chunk):
#     return (min(chunk), max(chunk))

def main():

    

    # print('Reading through video serially with no processing...')
    # takes 30 sec, (~500 it/s)
    # start_time = time.time()
    # with my_io.videoReader(test_movie, np.arange(500, 700)) as vid:
    #     for frame in tqdm(vid):
    #         pdb.set_trace()
    #         _ = dummy_process_frame(frame)
    # end_time = time.time()
    # dt = end_time - start_time
    # print(f'Dummy serial processing took {dt} seconds ({nframes/dt} it/s)')


    # time how long it takes just to read through the video in parallel
    # (This is pickling the frames before they go out to the workers)
    # takes 166 seconds, 180 it/s
    # print('Reading through video in parallel with no processing...')
    # start_time = time.time()
    # with my_io.videoReader(test_movie) as vid:
    #     process_map(dummy_process_frame, vid, chunksize=500, max_workers=6)
    # end_time = time.time()
    # dt = end_time - start_time
    # print(f'Dummy parallel processing took {dt} seconds ({nframes/dt} it/s)')

    # time how long it takes to use a generator object with vidReader and dispatch tqdm parallel it
    # takes 177 seconds, 95 of which are processing and remainder is setup/teardown.
    # runs at 190 it/s. 
    # print('Processing frames in parallel...')
    # start_time = time.time()
    # with my_io.videoReader(test_movie) as vid:
    #     process_map(process_frame, vid, chunksize=500, max_workers=6)
    # end_time = time.time()
    # print(f'Parallel processing with single reader gen took {end_time - start_time} seconds')


    # time how long it takes to get a bunch of vidReaders on the same video and have them each read a frame chunk
    # substantially slower, each one goes at about 100 it/s, but they seem to run serially!
    # all readers is started up front, but then you only get one going at a time.
    # print('Sending parallel workers to read video separately...')
    # max_workers = 6
    # frame_ixs = list(map(np.asarray, gen_batch_sequence(nframes, nframes//max_workers, 0)))
    # start_time = time.time()
    # process_map(parallel_dummy_process_frame, repeat(test_movie), frame_ixs, count(0), max_workers=max_workers)
    # end_time = time.time()
    # dt = end_time - start_time
    # print(f'Parallel processing with multiple readers took {end_time - start_time} seconds ({nframes/dt} it/s)')

    # print('Testing chunk size arg...')
    # # data = list(np.arange(1000))  # Ah, this doesn't work! The iterables are chunked for pickling, but they're still iterated individually!
    # frame_chunksize = 500
    # data = list([np.arange(i,i+frame_chunksize) for i in range(0,2000, frame_chunksize)])  # this is what we want
    # output = process_map(chunksize_test_func, data, chunksize=1, max_workers=4)
    # print(output)

    # finally, do what we're actually interested in: 
    # make a bunch of video writers, write individual videos from each process, and zip them together at the end
    # takes 60 seconds! f**k yeah. Now we hope that stitching is fast...
    # print('Reading through video in parallel and writing many outputs...')
    # start_time = time.time()
    # max_workers = 3
    # frame_chunksize = 500
    # batch_seq = gen_batch_sequence(nframes, frame_chunksize, 0)
    # parallel_writer_names = [f'./out/proc_{i}.avi' for i in range(len(batch_seq))]
    # # reporter_vals = (i for i in range(len(batch_seq)))
    # reporter_vals = (None for i in range(len(batch_seq)))
    # process_map(parallel_writing_process_frame, repeat(test_movie), batch_seq, reporter_vals, parallel_writer_names, chunksize=1, max_workers=max_workers)
    # end_time = time.time()
    # dt = end_time - start_time
    # print(f'Processing took {dt} seconds ({nframes/dt} it/s)')


    # debugging: show first frame of each parallel vid, and each supposed corresponding frame of full vid
    # these match perfectly. so the issue is in the stitching.
    # for batch, out_name in zip(batch_seq, parallel_writer_names):
    #     frame_num = np.array(int(batch.start))
    #     relative_frame_num = np.array([0])
    #     with my_io.videoReader('./out/proc.avi', frame_num) as full_vid, \
    #          my_io.videoReader(out_name, relative_frame_num) as stub_vid:        
    #          for full_frame, stub_frame in zip(full_vid, stub_vid):
    #             # plt.subplot(1,2,1)
    #             # plt.imshow(full_frame)
    #             # plt.subplot(1,2,2)
    #             # plt.imshow(stub_frame)
    #             # plt.suptitle(f'Frame {frame_num}')
    #             # plt.savefig(f'./out/debug_full_vs_stitched_frame_{frame_num}.png')

    #             if not np.all(full_frame == stub_frame):
    #                 print(f'Frame {frame_num} not equal in serial and stub videos')


    # stich the videos back together. (takes 30 seconds)
    # print('Stitching parallel videos')
    # out_vid = './out/stitched_proc.avi'
    # start_time = time.time()
    # with my_io.videoWriter(out_vid) as writer:
    #     for i, vid_name in enumerate(parallel_writer_names):
    #         print(f'Stitching video {i}')
    #         with my_io.videoReader(vid_name) as vid:
    #             for frame in vid:
    #                 writer.append(frame)
    # end_time = time.time()
    # dt = end_time - start_time
    # print(f'Processing took {dt} seconds ({nframes/dt} it/s)')



    print('Validating stitched video...')
    serial_name = './out/proc.avi'
    stitched_name = './out/stitched_proc.avi'
    
    print(f'Serial mov has {get_reader(serial_name).count_frames()} frames')
    print(f'Stitched mov has {get_reader(stitched_name).count_frames()} frames')

    equal_frames = np.zeros(get_reader(serial_name).count_frames())
    non_equal_counter = 0
    print_every_n_nonequal = 1000
    with io.videoReader(serial_name) as serial_vid, io.videoReader(stitched_name) as stitched_vid:
        for iFrame, (serial_frame, stitched_frame) in enumerate(zip(serial_vid, stitched_vid)):
            if not np.all(serial_frame == stitched_frame):
                print(f'Frame {iFrame} not equal in serial and stitched videos')
                non_equal_counter += 1

                print('Looking for matching frames...')

                # if non_equal_counter % print_every_n_nonequal == 0:
                #     plt.subplot(1,2,1)
                #     plt.imshow(serial_frame)
                #     plt.subplot(1,2,2)
                #     plt.imshow(stitched_frame)
                #     plt.suptitle('Frame {iFrame}')
                #     plt.savefig(f'./out/non_equal_frames_{iFrame}.png')
            else:
                equal_frames[iFrame] = 1
    
    plt.figure()
    plt.plot(equal_frames)
    plt.savefig('./out/equal_frames.png')

    if non_equal_counter == 0:
        print('Congratulations!! The videos match.')

    


    

if __name__ == '__main__':
    main()


import io
import re
from glob import glob
from os.path import basename
from os.path import dirname
from os.path import join
from os.path import splitext

from setuptools import find_packages
from setuptools import setup


setup(
    name='videochef',
    version='0.0.1',
    license='MIT',
    description='Whip up some fast, embarrassingly parallel video processing',
    author='Jonah Pearl',
    author_email='jonahpearl@g.harvard.edu',
    url='https://github.com/jonahpearl/videochef',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    py_modules=[splitext(basename(path))[0] for path in glob('src/*.py')],
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        # complete classifier list: http://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 1 - Planning',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Operating System :: Unix',
        'Operating System :: POSIX',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering :: Image Processing',
    ],
    project_urls={

        'Issue Tracker': 'https://github.com/jonahpearl/videochef/issues',
    },
    python_requires='>=3.6',
    install_requires=[
        'numpy', 'click', 'av', 'tqdm',
    ],
)